######################################################################
#:CQSchema.pm
#
# Licensed Materials - Property of IBM
# Rational ClearCase
# (C) Copyright IBM Corp. 2001, 2008 All Rights Reserved
# US Government Users Restricted Rights
# Use, duplication or disclosure restricted by GSA ADP Schedule Contract with IBM Corp.
#
# Copyright 2001,2008 Rational Software Corporation. All Rights Reserved
# This software contains proprietary and confidential information of
# Rational and its suppliers.  Use, disclosure or reproduction is
# prohibited without the prior express written consent of Rational.

=head1 NAME

CQCC::CQSchema - Class used to manage ClearQuest schema descriptions

=head1 CLASS DESCRIPTION

This class manages schema descriptions to support different schemas
across multiple entity record types. As used here a schema defines 
a generic record type consisting of a field map (mapping between
the default names Rational ClearQuest initially uses and the local
names for the same fields), one or more query descriptions (defining 
the search filters) and one or more results descriptions (defining
the fields being returned and the desired display format string). 

SUPPORT POLICY: See TriggerCQCC.pm for the Rational Support Policy. 

=head1 METHODS

=cut

######################################################################
#::Package CQCC:CQSchema
######################################################################

package CQCC::CQSchema;
use Exporter;
@ISA = ('Exporter');
@EXPORT = qw(&DefineCQSchema &GetCQSchemaByEntity 
             &SupportedDatabaseList &SupportedEntityList
             &GetRestrictionsQuery );

use strict;
sub Version { return '2.1' };
use CQCC::IOBase;
use CQCC::Parm;

######################################################################
#::Class Globals 
######################################################################

my %CQSchemaByEntity;     # Hash holding defined CR CQSchema objects 
my %CQSchemaByName;       # Hash holding all schema objects
my %CQSchemaByDatabase;   # Hash holding schemas by database

######################################################################
#::function ConvertValues
# Handles conversion and error reporting

sub ConvertValues {
    my ($fieldname, @values) = @_;
    # Use ParmIOConvertString since it will not cause IOBase to be 
    # needlessly created while we are still in config file processing
    my ($status,@convertedValues) = 
        &CQCC::Parm::ParmIOConvertString(\@values,"UTF8_CODEPAGE","CORE_CODEPAGE");
    if ($status) {
        IOMessage("FATAL", ["CQS_E_CONVERSION_ERROR",$fieldname,$status]);
    }
    return @convertedValues;
}

######################################################################
#::method ChangeFieldMap

=over 6

=item C<ChangeFieldMap(key=>value, ...)>

This method adds or overrides field mappings in the field map. The
standard mappings will persist unless they are directly overwritten.

=back

=cut 

######################################################################

sub ChangeFieldMap { 
    my ($self, @changes) = @_;

    my %fieldMap = %{ $self->{FIELD_MAP} };
    my ($key, $value);
    
    while ($key = shift @changes) {
        $value = shift @changes;
        $fieldMap{$key} = $value;
    }
    $self->{FIELD_MAP} = \%fieldMap;
}

######################################################################
#::function ClassInit

=over 6

=item C<ClassInit()>

This class function is automatically invoked to define class 
parameters using CQCC::Parm objects. Current parameters include: 

=over 4

=item C<CQCC_DATABASE_ENTITY_LIST>

This parameter defines one or more databases and which entities they
support. The format is

'dbname1: entity1,entity2; $dbname2: entity3,entity4'

i.e. 'SAMPL: defect,patch; SAMPL2: patch,defect'

=back

=back

=cut

######################################################################

sub ClassInit {
    &SetParmDefaults(PCLASS=>"CQCC::CQSchema", SRC_CFG=>1);

    &DefineParm(NAME=>"CQCC_DATABASE_ENTITY_LIST", 
                DEFAULT=>"SAMPL: defect");

    &SetParmDefaults();
}

&ClassInit();

######################################################################
#::function ClassTest

=over 6

=item C<ClassTest(@args)>

This function will create schema samples, lookup schemas for entities,
and return various values. 

=back

=cut

######################################################################

sub ClassTest { 
    my @args = @_;

    ############### Sample Schema definition
    my $s = &DefineCQSchema(NAME=>"Schema1",
                            ENTITY_LIST => "defect, patch",
                            RESTRICTIONS => "STANDARD");
    $s->ChangeFieldMap(OWNER => "Owner",
                      STATE => "State",
                      ID => "id",
                      QUERY_STATES => "Submitted,Assigned,Opened"
                     );
    $s->SetQuery("STANDARD", 
                 "<owner> eq <*user*>",
                 "<STATE> in Open,Closed VeryMuddled");
    $s->SetResultSet("STANDARD",
                     "<ID>,<STATE> <Priority>, Severity, Headline",
                     "%s %-9.9s %1.1s %1.1s %-45.45s");

    ############### Sample Schema definition (2)
    $s = &DefineCQSchema(NAME=>"Schema2",
                         ENTITY_LIST => "docfix escalation");
    $s->ChangeFieldMap(OWNER => "AssignedTo",
                       STATE => "Massachusetts",
                       ID => "number",
                       QUERY_STATES => "Confused,Angry,Befuddled"
                       );
    $s->SetQuery("STANDARD", 
                 "<owner> eq <*user*>",
                 "<STATE> in Open,Closed VeryMuddled");
    $s->SetResultSet("STANDARD",
                     "<ID>,<STATE> <Priority>, Severity, Headline",
                     "%s %-9.9s %1.1s %1.1s %-45.45s");
    
    ############### Other Variables

    &ForceParm("CQCC_DATABASE_ENTITY_LIST",
         "SAMPL: defect, docfix; SAMPL2: defect, patch, escalation");

    ############### Retrieve information

    print "SupportedDatabaseList: " .
        join(",", &SupportedDatabaseList()) . "\n";

    print "SupportedEntityList: " .
        join(",", &SupportedEntityList("SAMPL")) . "\n";

    $s->Print();
    $s->Print("translate");

    print "Test translation of one-time query\n";
    my @filters = $s->TranslateQuery(["<STATE> in <QUERY_STATES>"]);
    print join("\n", @filters) . "\n";
    return 0;
}

######################################################################
#::function DefineCQSchema 

=over 6

=item C<DefineCQSchema(key=>value, ...)>

This function defines a new schema and adds it to the global list. 
It also defines the standard field_map, query_list, results_set
and results_format information for a ClearQuery change request
record. 

=back

=cut 

######################################################################

sub DefineCQSchema { 
    my @args = @_;
    my $schema = new CQCC::CQSchema(TYPE=>"CR");
    my %validKeys = (NAME=>1, ENTITY_LIST=>1, RESTRICTIONS=>1);
    my %fieldMap = (
        HEADLINE => "Headline",
        ID => "id",      
        OWNER => "Owner",        
        PRIORITY => "Priority",  
        SEVERITY => "Severity",
        STATE => "State",
        RATL_MASTERSHIP => "ratl_mastership",

        # Other tokens that need to be changeable
        CC_CHANGE_SET => "cc_change_set",
        QUERY_STATES => "Submitted,Assigned,Opened",
        MODIFY => "modify"
    );

    my %queryList = (
        "STANDARD" => ["<OWNER> eq <*USER*>",
                       "<STATE> in <QUERY_STATES>" ],
    );

    my %resultsSetList = (
        "STANDARD" =>
            "<ID>, <STATE>, <PRIORITY>, <SEVERITY>, <HEADLINE>",
    );

    my %resultsFormatList = (
        "STANDARD" => "%s %-9.9s %1.1s %1.1s %-45.45s",
    );

    $schema->{FIELD_MAP} = \%fieldMap;
    $schema->{QUERY_LIST} = \%queryList; 
    $schema->{RESULTS_SET_LIST} = \%resultsSetList;
    $schema->{RESULTS_FORMAT_LIST} = \%resultsFormatList;

    while (@args) { 
        my $key = shift @args;
        my $value = shift @args;
        if ($validKeys{$key}) {
            $schema->{$key} = $value;
        }
        else {
            &IOMessage("ERROR", 
                [ "CQS_E_UNKNOWN_KEY", $key, $schema->{NAME}]);
        }
    } 

    # Store references to this schema object by each supported entity
    my @entlist = $schema->GetEntityList();
    for (@entlist) { 
        $CQSchemaByEntity{$_} = $schema;
    }

    if (defined $schema->{NAME}) {
        $CQSchemaByName{$schema->{NAME}} = $schema;
    } 
    return $schema;
}

######################################################################
#::method EvalItem

=over 6

=item C<EvalItem(item)>

This method evaluates it's argument to process embedded references to
the objects instance variables like <OWNER>.

=back

=cut 

######################################################################

sub EvalItem {
    my ($self, $item) = @_;
    my %fieldMap = %{ $self->{FIELD_MAP} };
    my $cquser;

    if ($item =~ m/\<(\S*)\>/) {
        my $token = uc($1);
        if ($token eq "*USER*") {
           if (&ParmDefined("CQUSER")) {
               $cquser = &GetParm("CQUSER");               
           } 
           if ( defined $cquser ) {
               $item = $cquser;
           } elsif (&ParmDefined("USER")) {
               $item = &GetParm("USER");
           } else {
               $item = "<*USER*>";
           }
        } elsif ($token eq "*REPLICA_NAME*") {
           if (&ParmDefined("REPLICA_NAME")) {
               $item = &GetParm("REPLICA_NAME");
           } else {
               $item = "<*REPLICA_NAME*>";
           }
        } else {
           if ($fieldMap{$token}) {
                $item = $fieldMap{$token};
           }
        }
    } 
    my ($itemsList) = &ConvertValues("EvalItem",$item);
    my @newItemsList = split(",",$itemsList);
    return @newItemsList;
}
    
######################################################################
#::function GetCQSchemaByEntity

=over 6

=item C<GetCQSchemaByEntity(entityname)>

This function looks up an existing schema that supports the specified
entity name. 

=back

=cut

######################################################################

sub GetCQSchemaByEntity {
    my $entity = shift;

    return $CQSchemaByEntity{$entity};
}
                    
######################################################################
#::function GetCQSchemaByName

=over 6

=item C<GetCQSchemaByName(schemaName)>

This function looks up an existing schema with the given name. 


=back

=cut

######################################################################

sub GetCQSchemaByName {
    my $name = shift;

    return $CQSchemaByName{$name};
}
                    
######################################################################
#::method GetEntityList

=over 6

=item C<GetEntityList()>

This method retrieves the entity list from the object after parsing
the original string. 

=back

=cut

######################################################################

sub GetEntityList {
    my $self = shift;
    my ($list) = &ConvertValues("EntityList",$self->{ENTITY_LIST});
    return &IOParseListString($list);
}
                    
######################################################################
#::method GetQuery

=over 6

=item C<GetQuery(name)>

This method looks up the provided name and retrieves the matching
query clauses as a list. 

=back

=cut

######################################################################

sub GetQuery {
    my ($self, $name) = @_;

    my %queryList = %{ $self->{QUERY_LIST} };
    my $queryEntry = $queryList{$name};
    if (defined $queryEntry) {
        return &ConvertValues("Query",@{ $queryEntry });
    }
    return;    
}
                    
######################################################################
#::method GetQueryList

=over 6

=item C<GetQueryList(name)>

This method returns the list of queries defined for the current
schema. 

=back

=cut

######################################################################

sub GetQueryList {
    my ($self, $name) = @_;
    my %queryList = %{ $self->{QUERY_LIST} };
    return &ConvertValues("Querylist",keys %queryList);
}
                    
######################################################################
#::method GetResultSet

=over 6

=item C<GetResultSet(name)>

This method looks up the provided name and retrieves the matching
resultSet string (comma delimited list) and the format string. If
any field mappings are embedded in the resultSet list the
substitutions are made before the result is returned.

($resultSet, $format) = $obj->GetResultSet(name);

If the name provided is an array it is treated as a one time result
set of [ result_set, format ]. The result_set will be translated and
returned. 

=back

=cut

######################################################################

sub GetResultSet {
    my ($self, $name) = @_;
    my ($resultSet, $resultFormat);

    if (ref($name) eq "ARRAY") {
        ($resultSet, $resultFormat) = @{ $name };
    } else {
       my %resultSetList = %{ $self->{RESULTS_SET_LIST} };
       my %resultFormatList = %{ $self->{RESULTS_FORMAT_LIST} };
       $resultSet = $resultSetList{$name};
       $resultFormat = $resultFormatList{$name};
    }
       
    my @resultSetRaw = &IOParseListString( $resultSet );
    my @resultSet;

    for (@resultSetRaw) { 
        push @resultSet, $self->EvalItem($_);
    }
    ($resultFormat) = &ConvertValues("ResultSet",$resultFormat);
    return (join(",",@resultSet), $resultFormat);
}
                    
######################################################################
#::method GetRestrictionsQuery

=over 6

=item C<GetRestrictionsQuery(entity)>

This method retrieves the name of the restrictions query associated
with the given entity if any.

=back

=cut

######################################################################

sub GetRestrictionsQuery {
    my ($self, $entity) = @_;
    if ($self->{RESTRICTIONS}) {
        my @query = $self->GetQuery($self->{RESTRICTIONS});
        if (@query) {
            # Note: ConvertValues already handled by GetQuery
            return @query;
        } else {
            &IOMessage("ERROR", 
                [ "CQS_E_UNKNOWN_QUERY", $self->{RESTRICTIONS},
                  $self->{NAME} ]);
        }
    }
    return ();
}
                    
######################################################################
#::method new

=over 6

=item C<new CQCC::CQSchema()>

This method defines a new schema and adds it to the global list.
It supports the following instance variables. 

=over 4 

=item C<NAME>

A name for the schema for easier reference. 

=item C<ENTITY_LIST>

A scalar string containing a comma delimited list of entity names
supported by this schema. At least one name must be provided. 

=item C<QUERY_NAME_PATTERN>

A scalar string defining the public named queries that can be 
presented to the user. Example "public queries\temp\" would select
all queries beginning with "public queries\temp\".

=item C<FIELD_MAP>

A hash table of field name mappings, supporting local schema changes.
Queries and results lists can use either explicit field names or
references to the field map lists. For example if OWNER=>AssignedTo
was in the field map, then <OWNER> would be translated to "AssignedTo"
before it was used. This allows the application to generate additional
queries on the fly. It will also facilitate taking advantage of 
ClearQuest schema aliases when that feature becomes available. 

=item C<QUERY_LIST>

A hash table containing named query strings allowing Schemas to 
support more than one standard query, i.e. "Standard" and 
"Restriction". Each query is an array of one or more clauses, each of 
which is in the form "subject verb object". Each clause will be 
processed to replace any <> references with the appropriate field 
mappings, allowing generic queries to be defined that work with 
different field mappings. 

=item C<RESULTS_SET_LIST>

A hash table containing named results set lists in the form of
scalar strings delimited with commas. When a query is performed
the results set defines which entity fields are to be retrieved. 
A matching format string with the same name is used to format
the columns into a return string for each record being retrieved. 
The default results set, "STANDARD", is set to this string: 
"<ID>, <STATE>, <PRIORITY>, <SEVERITY>, <HEADLINE>"

=item C<RESULTS_FORMAT_LIST>

A hash table containing named format strings to match the results
set with the same name. The default string is set to this string:
"%s %-9.9s %1.1s %1.1s %-45.45s".

=back 

=back

=cut

######################################################################

sub new { 
   my $proto = shift;
   my $class = ref($proto) || $proto;  # Object or class name

   my %emptyHash = ();
   my $e = \%emptyHash;

   # Attributes
   my $self = {
       # Required instance data
       NAME => undef,              # CQSchema Name
       CLASSNAME => "CQCC::CQSchema",
       TYPE => undef,              # Generic record type, CR, etc
       ENTITY_LIST => "",          # Entities supported by schema
       FIELD_MAP => $e,            # Field name aliases
       QUERY_LIST => $e,           # Named queries
       RESULTS_SET_LIST => $e,     # Results set lists
       RESULTS_FORMAT_LIST => $e,  # Matching formats for results
       RESTRICTIONS => undef,      # Name of Query to test against
       @_,                         # Override previous attributes
       @_,                         # Extra time for ccperl
   };


   # Add test that all required values (Name and defval)
   
   bless($self, $class);
   if (defined $self->{NAME}) {
      $CQSchemaByName{$self->{NAME}} = $self;
   } 
   return $self;
}

######################################################################
#::method Print

=over 6

=item C<Print()>

This method prints out the contents of the schema.

=back

=cut

######################################################################

sub Print {
    my ($self, $translated) = @_;

    my %fieldMap = %{ $self->{FIELD_MAP} };
    my %queryList = %{ $self->{QUERY_LIST} };
    my %resultsSetList = %{ $self->{RESULTS_SET_LIST} };
    my %resultsFormatList = %{ $self->{RESULTS_FORMAT_LIST} };

    &IOPrint( "Name: $self->{NAME}\nType:  $self->{TYPE}\n");
    &IOPrint( "Entities: " . join(",",$self->GetEntityList()) . "\n");
    &IOPrint( "Restrictions: $self->{RESTRICTIONS}\n");
    &IOPrint( "\nFieldMap: \n");
    for (sort keys %fieldMap) { 
        &IOPrint( "    $_ => $fieldMap{$_}\n");
    }
    &IOPrint( "\nQueryList: \n");
    for (sort keys %queryList) {
        my @clauses = $translated ? $self->TranslateQuery($_)
                                  : @{ $queryList{$_} };
        &IOPrint( "    $_ => \n");
        &IOPrint( "        " . join("\n        ",@clauses) . "\n");
    }
    &IOPrint( "ResultSets: \n");
    for (sort keys %resultsSetList) { 
        my ($resultSet, $format) = $self->GetResultSet($_);
        &IOPrint( "    $_ => " .
             ($translated ? $resultSet : $resultsSetList{$_})
             . "\n");
        &IOPrint( "        format: $format\n");
    }
    &IOPrint( "\n");
    return;
}

######################################################################
#::method SetQuery

=over 6

=item C<SetQuery(name, query_clauses...)>

This method adds or overrides query definitions in the schema.  Each
query needs a name and an array of conditional clauses which may
contain field aliases, i.e. "<OWNER> eq <*USER*>".  The standard
mappings will persist unless they are directly overwritten.

=back

=cut 

######################################################################

sub SetQuery { 
    my ($self, $name, @clauseArray) = @_;

    my %queryList = %{ $self->{QUERY_LIST} };
    $queryList{$name} = \@clauseArray;
    $self->{QUERY_LIST} = \%queryList;
}

######################################################################
#::method SetResultSet

=over 6

=item C<SetResultSet(name, resultSetList, resultFormat)>

This method adds or overrides result set descriptions in the schema
based on a comma delimited string with the desired fields and a
matching format string. The resultSetList may contain field aliases,
i.e. "<OWNER> eq <*USER*>".  The standard mappings will persist unless
they are directly overwritten.

=back

=cut 

######################################################################

sub SetResultSet { 
    my ($self,$name,$resultSet,$format) = @_;
    my %resultSetList = %{ $self->{RESULTS_SET_LIST} };
    my %resultFormatList = %{ $self->{RESULTS_FORMAT_LIST} };

    $resultSetList{$name} = $resultSet;
    $resultFormatList{$name} = $format;

    $self->{RESULTS_SET_LIST} = \%resultSetList;
    $self->{RESULTS_FORMAT_LIST} = \%resultFormatList;
}

######################################################################
#::function SupportedDatabaseList

=over 6

=item C<SupportedDatabaseList()>

This function returns a list of all supported databases supported by
the defined schemas. 

=back

=cut

######################################################################

sub SupportedDatabaseList {
    my $self = shift;
    my $dbEntStringList = &GetParm("CQCC_DATABASE_ENTITY_LIST");
    $dbEntStringList =~ s/\n//g;
    $dbEntStringList =~ s/ //g;
    my @dbEntList = split(";",$dbEntStringList);
    my @dbList;
    
    for (@dbEntList) {
       my ($db, $entityList) = split(":",$_);
       push @dbList, $db;
    }

    return &ConvertValues("SupportedDatabaseList", @dbList);
}
                    
######################################################################
#::function SupportedEntityList

=over 6

=item C<SupportedEntityList(databaseName)>

This function gets entities supported by a specific database.

=back

=cut

######################################################################

sub SupportedEntityList {
    my $database = shift;
    my $dbEntStringList = &GetParm("CQCC_DATABASE_ENTITY_LIST");
    $dbEntStringList =~ s/\n//g;
    $dbEntStringList =~ s/ //g;
    my @dbEntList = split(";",$dbEntStringList);
    my @entList = ();
    
    for (@dbEntList) {
       my ($db, $entityList) = split(":",$_);
       if ($db eq $database) {
           $entityList =~ s/,/ /g;
           $entityList =~ s/  / /g;
           @entList = split(" ", $entityList);
           # Note: Conversion already handled by GetParm
           return @entList;
       } 
    }
    return @entList;
}
                    
######################################################################
#::method TranslateQuery

=over 6

=item C<TranslateQuery(queryName or [ clauselist ] )>

This method takes either the name of a defined query or a one-time
query provided as a list of query clauses. In either case it will
process the query clauses to substitute field mapping names in 
in place of the <NAME> tags. A special tag, <*USER*> is provided to
reference the current USER name in the midst of the current trigger
invocation.

=back

=cut 

######################################################################

sub TranslateQuery {
    my ($self, $queryName) = @_;
    my @clauseList;
    my $qName;

    if (ref($queryName) eq "ARRAY") {   # If query name is undef then
        #Note: runtime query assumed to be CORE_CODEPAGE already
        @clauseList = @{ $queryName }; # get clause list from args
        
        $qName = "INTERNAL";
    } else {
        # Note: GetQuery handles conversion
        @clauseList = $self->GetQuery($queryName);
        $qName = $queryName;
    }

    my @newClauses;
    my ($clause, $subject, $verb, $valueString, $value, @args);
    
    for $clause (@clauseList) {
       # Parse pattern: token<whitespace>token<whitespace>restofline
      ($subject, $verb, $valueString) = split(" ", $clause, 3);
      if (defined $valueString && $valueString ne "") {
           my ($newSubject) = $self->EvalItem($subject);
           my @args = $self->EvalItem($valueString);
           my $arglist = join(",", @args);
           my $newClause = 
              join(" ",($newSubject, $verb, $arglist));
           push @newClauses, $newClause;
       } else {
           &IOMessage("FATAL", ["CQP_E_QUERY_INVALID", "[$clause]",
                   $qName . "\n" . join("\n  ",@clauseList)]);
       }
    }

    # Note: Earlier conversion assumed
    return @newClauses;
}

######################################################################
#::Tail
######################################################################

1;
